<?php
/*
//----------------------------------------------------------------------------
// Copyright (c) 2006-2011 Asymmetric Software - Innovation & Excellence
// Author: Mark Samios
// http://www.asymmetrics.com
// Admin: Nodes core class
// Loads/Saves node files, determines best paths among nodes, 
// using distances and vertices
// Supports loading and storing path files
// Multiple controls for shortest paths vs performance added
//----------------------------------------------------------------------------
// I-Metrics CMS
//----------------------------------------------------------------------------
// Script is intended to be used with:
// osCommerce, Open Source E-Commerce Solutions
// http://www.oscommerce.com
// Copyright (c) 2003 osCommerce
//----------------------------------------------------------------------------
// Released under the GNU General Public License
//----------------------------------------------------------------------------
*/
  class nodes {
    static $marked_array = array();
    static $nodes_array = array();
    static $verts_array = array();

    //var $optimum_array;
    var $stats_array, $control_array, $display_array, $path_result_array, $org_result_array;
    var $min_dist, $max_depth, $max_results, $max_visited, $max_node_connects, $debug_display;
    var $org_from, $org_to, $nearest;

    // class constructor
    function nodes($nodes='', $verts='', $debug_display=0) {
      $this->reset();

      if( !empty($nodes) ) {
        if( is_array($nodes) ) {
          nodes::$nodes_array = $nodes;
          if( !empty($verts) && is_array($verts) ) {
            nodes::$verts_array = $verts;
          }
        } else {
          nodes::load_path_file($nodes);
        }
      }
      $this->debug_display = $debug_display;
    }

    function reset($all=false) {
      if( $all ) nodes::reset_statics();

      // Statistics storage
      $this->stats_array = array(
        'recursions' => 0, // Regular Recursions occured
        'revisited_segments' => 0, // Optimized scans via marked_array
        'empty_visits' => 0, // Visits with Empty intermediate nodes skipped
        //'empty_intermediates' => 0, // Empty intermediate nodes skipped
        'revisited_nodes' => 0, // Already visited from nodes
        'distance_breaks' => 0, // Distance Breaks
        'distance_segment' => 0, // Distance Breaks
        'cache_hits' => 0, // Cache hits in marked_array
        'stack_peak' => 0, // Stack Peak from recursions
        'path_optimizations' => 0, // Path Optimizations
      );

      // Optimization controls
      $this->control_array = array(
        'use_vertices' => 1, // 1=Priority pass via Vertices, 0=priortity via Distance among nodes
        'distance_breaks' => 1, // 1=Enable Parent Distance Breaks (note: its mutually exclusive with distance_segment)
        'distance_segment' => 1, // 1=Enable Distance Segment Breaks
        'path_optimizations' => 1, // 1=Enable Path Optimizations
        'stack_peak' => 0, // 1=Enable Stack Control for broad scans and huge node sets
        'results_peak' => 0, // 1=Break out if max resulted paths is reached
      );

      // Additional display details
      $this->display_array = array(
        'nodes_array' => 1,
        'verts_array' => 0,
        'marked_array' => 0,
      );

      $this->nearest = 0;
      $this->max_node_connects = 4;
      $this->max_depth = 4;
      $this->path_result_array = $this->org_result_array = array();
      $this->set_max_visited();
      $this->set_max_results();
    }

    static function reset_statics() {
      nodes::$marked_array = nodes::$nodes_array = nodes::$verts_array = array();
    }

    function set_stack_depth($depth=4) {
      $this->max_depth = $depth;
    }

    function set_max_results($max=5) {
      $this->max_results = $max;
    }

    function set_max_visited($max=0) {
      if( empty($max) ) {
        $this->max_visited = $this->max_depth*$this->max_node_connects;
      } else {
        $this->max_visited = $max;
      }
    }

    function get_path($from, $to) {
      $this->org_from = $from;
      $this->org_to = $to;
      $tmp_array = nodes::$nodes_array;
      $this->get_path_recursive($from, $to);
      nodes::$nodes_array = $tmp_array;
    }

    function get_path_recursive($from, $to, $visited=array(), $current_array=array() ) {

      $depth = count($current_array);
      $min_dist = array_sum($current_array);
      $current_visits = count($visited)-$depth;

      if( $this->org_from == $from ) {
        $this->set_max_visited();
      }

      $connected = array();

      if( !isset(nodes::$nodes_array[$from]) || !isset(nodes::$nodes_array[$to]) || $from == $to ) {
        if( $this->debug_display ) {
          echo sprintf('Node connection does not exist from %s to %s - Returning', $from, $to) . '<br />';
        }
        return $connected;
      }

      if( $this->control_array['results_peak'] && count($this->path_result_array) >= $this->max_results ) {
        return $connected;
      }

      if( $this->control_array['stack_peak'] && $depth > $this->max_depth ) {
        return $connected;
      }

      if( $this->stats_array['stack_peak'] < $depth ) {
        $this->stats_array['stack_peak'] = $depth;
      }

      $this->set_nearest($from, $to);

      if( $this->check_branch($from, $to, $visited, $current_array) ) {
        return $connected;
      }

      if( !isset(nodes::$nodes_array[$from][$to]) ) {

        $visited[$from] = $from;

        if( $current_visits > $this->max_visited ) {

          if( $this->debug_display ) {
            echo sprintf('Visited Limit reached from %s to %s Limit=%s - Aborting', $from, $to, $this->max_visited) . '<br />';
          }
          return $connected;
        }

        if( $this->is_distance_break($from, $to, $visited, $current_array) ) {
          return $connected;
        }

        $visited_tmp_array = $visited+nodes::$nodes_array[$from];

        $nodes_array = $this->short_vertices(nodes::$nodes_array[$from], $to);

        foreach( $nodes_array as $key => $value ) {

          $this->set_reverse($from, $key, $to);

          if( $this->is_visited($from, $key, $to, $visited) ) {
            continue;
          }

          $current_tmp_array = $current_array;
          $current_tmp_array[$from] = $value;

          if( $this->is_distance_segment_break($from, $key, $to, $current_tmp_array) ) {
            return $connected;
          }

          if( isset(nodes::$marked_array[$from][$key]) && empty(nodes::$marked_array[$from][$key]) ) {
            $this->check_current_segment($from, $key, $current_tmp_array);
          }

          if( !isset(nodes::$marked_array[$from][$key]) ) {
            $this->stats_array['recursions']++;
            if( $this->debug_display ) {
              echo sprintf('Trying Path [%s] via %s to %s', implode(',', array_keys($current_tmp_array)), $key, $to) . '<br />';
              //echo sprintf('Visited are [%s]', implode(',', array_keys($visited_tmp_array)) ) . '<br />';
            }
            $tmp_array = $this->get_path_recursive($key, $to, $visited_tmp_array, $current_tmp_array);
            if( !empty($tmp_array) ) {
              // For First match return result from here
              // return $tmp_array;
            } else {
              if( $this->debug_display ) {
                echo sprintf('Blocking Path %s:[%s]', $from, $key) . '<br />';
              }
              $this->set_marked($from, $key, array());
            }

            //if( empty($tmp_array) ) {
            //  $this->stats_array['empty_intermediates']++;
            //  if( $this->debug_display ) {
            //    echo sprintf('Empty intermediate set from %s via %s to %s', $from, $key, $to) . '<br />';
            //  }
            //}
            continue;

          } else {

            if( empty(nodes::$marked_array[$from][$key]) ) {
              $this->stats_array['empty_visits']++;
              if( $this->debug_display ) {
                echo sprintf('Revisit with empty intermediate with [%s] via %s to %s', implode(',', array_keys($current_tmp_array)), $key, $to ) . '<br />';
              }
              continue;
            }

            $this->stats_array['revisited_segments']++;
            if( $this->debug_display ) {
              echo sprintf('Path Already Parsed [%s] via %s to %s', implode(',', array_keys($current_tmp_array)), $key, $to) . '<br />';
            }

            $tmp_array = tep_array_splice_from_key(nodes::$marked_array[$from][$key], $current_tmp_array, $from);
            end($tmp_array);
            $last_key = key($tmp_array);

            $dist = array_sum($tmp_array);
            $path_key = $this->get_path_key($tmp_array, $dist);

            if( $this->is_distance_segment_break($from, $key, $to, $tmp_array) ) {
              if( $last_key == $to ) {
                $this->org_result_array[$path_key] = $tmp_array;
                $this->optimize_marked();
              }
              return $tmp_array;
            }

            if( $last_key == $to ) {

              $this->stats_array['cache_hits']++;
              if( $this->debug_display ) {
                echo sprintf('<b>Cached Segment Hit [%s]</b>', implode(',', array_keys($tmp_array)) ) . '<br />';
              }

              $this->org_result_array[$path_key] = $tmp_array;
              $this->path_result_array[$path_key] = $tmp_array;
              ksort($this->path_result_array, SORT_NUMERIC);
              $this->optimize_marked();
              $this->reduce_visits();
              return $tmp_array;
            }
          }
        }

        if( !empty($this->path_result_array) ) {
          foreach( nodes::$nodes_array[$from] as $key => $value ) {
            if( isset(nodes::$marked_array[$from][$key]) && empty(nodes::$marked_array[$from][$key]) ) {
              if( $this->debug_display ) {
                echo sprintf('Releasing Path %s:[%s]', $from, $key ) . '<br />';
              }
              unset(nodes::$marked_array[$from][$key]);
              if( empty(nodes::$marked_array[$from]) ) {
                unset(nodes::$marked_array[$from]);
              }
            }
          }
        }

      } else {
        $current_array[$from] = nodes::$nodes_array[$from][$to];
        $current_array[$to] = 0;

        $connected = $current_array;
        $dist = array_sum($connected);

        $path_key = $this->get_path_key($connected, $dist);

        $tmp_array = array();
        if( $this->is_distance_break($from, $to, $tmp_array, $current_array) ) {
          $this->org_result_array[$path_key] = $connected;
          $this->optimize_marked();
          if( $this->debug_display ) {
            echo sprintf('<b>Found Secondary Path [%s]</b>', implode(',', array_keys($connected)) ) . '<br />';
          }
          return $connected;
        }

        if( $this->debug_display ) {
          echo sprintf('<b>Found Main Path [%s]</b>', implode(',', array_keys($connected)) ) . '<br />';
        }
        $this->org_result_array[$path_key] = $connected;
        $this->path_result_array[$path_key] = $connected;
        ksort($this->path_result_array, SORT_NUMERIC);
        $this->optimize_marked();
        $this->reduce_visits();
      }
      return $connected;
    }

    static function get_node_ds($from, $to, $min_dist=1, $max_dist=4) {

      $result = 0;

      if( empty(nodes::$verts_array[$from]) ) {
        $result = rand($min_dist, $max_dist);
        return $result;
      }

      if( !isset(nodes::$verts_array[$from]) || !isset(nodes::$verts_array[$to]) || $from == $to ) {
        return $result;
      }

      $c = nodes::$verts_array[$from];
      $t = nodes::$verts_array[$to];

      $x = (abs($c['x']-$t['x']));
      $y = (abs($c['y']-$t['y']));
      $z = (abs($c['z']-$t['z']));

      $result = pow( $x*$x+$y*$y+$z*$z, 0.5);

      //$result = pow(pow($c['x']-$t['x'], 2)+pow($c['y']-$t['y'], 2)+pow($c['z']-$t['z'], 2), 0.5);
      return ceil($result);
    }

    function short_vertices($input_array, $to) {
      if( !$this->control_array['use_vertices'] ) {
        return $input_array;
      }

      $result_array = $tmp_array = array();
      $i = 0;
      foreach($input_array as $key => $value) {
        $i++;
        $d = nodes::get_node_ds($key, $to);
        $tmp_array[$d . '_' . $i] = $key;
      }
      ksort($tmp_array, SORT_NUMERIC);

      foreach($tmp_array as $value) {
        $result_array[$value] = $input_array[$value];
      }
      return $result_array;
    }

    function get_path_key($input_array, $dist) {
      $key = $dist . '_' . implode('_', array_keys($input_array));
      return $key;
    }

    function set_nearest($from, $to) {
      if( !$this->control_array['use_vertices'] ) return;

      $dist = nodes::get_node_ds($from, $to);
      if( empty($this->nearest) || $dist < $this->nearest['d'] ) {
        $this->nearest = array(
          'n' => $from,
          'd' => $dist
        );
      }
    }

    function set_reverse($from, $key, $to) {
      if( !$this->control_array['use_vertices'] || !isset(nodes::$nodes_array[$key][$from]) ) return false;

      $dfrom = nodes::get_node_ds($from, $to);
      $dkey = nodes::get_node_ds($key, $to);
      if( $dfrom > $dkey ) {
        unset(nodes::$nodes_array[$key][$from]);
      }
    }

    function clear_reverse($input_array) {
      $previous = 0;
      foreach($input_array as $ck => $cv ) {
        if( $previous && isset(nodes::$nodes_array[$ck][$previous]) ) {
          unset(nodes::$nodes_array[$key][$from]);
        }
        $previous = $ck;
      }
    }

    function set_marked($from, $to, $input_array) {
       nodes::$marked_array[$from][$to] = $input_array;
    }

    function clear_marked($input_array) {
      foreach($input_array as $key => $value) {
        unset(nodes::$marked_array[$key]);
      }
    }

    function check_branch(&$from, $to, &$visited, &$current_array) {
      $result = false;

      while( isset(nodes::$nodes_array[$from]) && count(nodes::$nodes_array[$from]) == 1 && !isset(nodes::$nodes_array[$from][$to]) ) {
        reset(nodes::$nodes_array[$from]);
        list($ck, $cv) = each(nodes::$nodes_array[$from]);
        $visited[$from] = $from;
        $current_array[$from] = nodes::$nodes_array[$from][$ck];

        if( $this->is_visited($from, $ck, $to, $visited) ) {
          $result = true;
          break;
        }

        if( $this->debug_display ) {
          echo sprintf('Single Connect Advance %s:[%s]', $from, $ck) . '<br />';
        }

        $from = $ck;
      }
      return $result;
    }

    function check_current_segment($from, $to, $current_array) {
      $result = false;
      if( !$this->control_array['path_optimizations'] ) return $result;

      $org_result_array = $this->org_result_array;

      foreach($org_result_array as $org_key => $value ) {
        if( isset($value[$from]) && isset($value[$to]) ) {

          end($current_array);
          $key = key($current_array);

          $new_array = tep_array_splice_from_key($value, $current_array, $key);
          if( count($new_array) < count($value) && array_sum($new_array) <= array_sum($value) ) {
            $this->stats_array['path_optimizations']++;
            if( $this->debug_display ) {
              echo sprintf('Re-optimizing path [%s] into [%s]', implode(',', array_keys($value)), implode(',', array_keys($new_array))) . '<br />';
            }
            $previous = 0;
            foreach($new_array as $ck => $cv) {
              if( $previous && isset(nodes::$marked_array[$previous][$ck]) ) {
                unset(nodes::$marked_array[$previous][$ck]);
                if( empty(nodes::$marked_array[$previous]) ) {
                  unset(nodes::$marked_array[$previous]);
                }
              }
              $previous = $ck;
            }

            unset($this->path_result_array[$org_key]);
            //unset($this->org_result_array[$org_key]);

            $dist = array_sum($new_array);
            $path_key = $this->get_path_key($new_array, $dist);

            if( !isset($this->path_result_array[$path_key]) ) {
              $this->path_result_array[$path_key] = $new_array;
              $this->org_result_array[$path_key] = $new_array;
              $result = true;
            }
          }
        }
      }

      if( $result ) {
        ksort($this->path_result_array, SORT_NUMERIC);
        $this->optimize_marked();
      }
      return $result;
    }

    function optimize_marked($path_array = array()) {
      nodes::$marked_array = array();

      foreach($this->org_result_array as $key => $value) {
        $previous = 0;
        foreach($value as $ck => $cv) {
          if( $previous ) {
            nodes::$marked_array[$previous][$ck] = array();
            // Clear reversed connections
            if( isset(nodes::$nodes_array[$ck][$previous]) ) {
              unset(nodes::$nodes_array[$ck][$previous]);
            }
          }
          $previous = $ck;
        }
      }

      reset($this->path_result_array);
      $tmp_array = current($this->path_result_array);

      if( empty($path_array) && !empty($tmp_array) ) {
        $path_array = $tmp_array;
      }

      $previous = 0;
      foreach($path_array as $ck => $cv) {
        if( $previous ) {
          nodes::$marked_array[$previous][$ck] = $path_array;
        }
        $previous = $ck;
      }
    }

    function reduce_visits($depth=1) {
      reset($this->path_result_array);
      $tmp_array = current($this->path_result_array);
      $factor = count($tmp_array)-$depth+$this->max_node_connects;
      if( $factor < $this->max_visited ) {
        $this->set_max_visited($factor);
        if( $this->debug_display ) {
          echo sprintf('Updating visited limit to %s', $factor) . '<br />';
        }
      }
    }

    function is_visited($from, $via, $to, $visited=array() ) {
      $result = false;
      if( isset($visited[$via]) || $from == $via ) {
        $this->stats_array['revisited_nodes']++;
        if( $from == $via ) {
          echo sprintf('Warning: Self-Connected node:[%s]', $from ) . '<br />';
        }
        if( $this->debug_display ) {
          //echo sprintf('Skipped %s via %s to %s as visited [%s]', $from, $via, $to, implode(',', array_keys($visited)) ) . '<br />';
          echo sprintf('Skipped %s via %s to %s as visited', $from, $via, $to) . '<br />';
        }
        $result = true;
      }
      return $result;
    }

    function is_distance_break($from, $to, &$visited, $current_array) {
      $result = false;
      if( !$this->control_array['distance_breaks'] || empty($this->path_result_array) ) return $result;

      $min_dist = array_sum($current_array);

      foreach($this->path_result_array as $ck => $cv) {
        $tmp_array = explode('_', $ck);
        $ds = $tmp_array[0];

        if( $ds <= $min_dist && count($cv) <= count($current_array) ) {
          $result = true;
        }

        if( !$result && !empty($visited) ) {
          foreach(nodes::$nodes_array[$from] as $key => $value) {
            if( $min_dist+$value >= $ds ) {
              $this->stats_array['distance_breaks']++;
              $visited[$key] = $key;
              if( $this->debug_display ) {
                echo sprintf('Distance Break d(%s) <= d(%s) setting visited from %s to %s', $ds, $min_dist+$value, $from, $key) . '<br />';
              }
            }
          }
        }
        break;
      }

      if( $result ) {
        $this->stats_array['distance_breaks']++;
        if( $this->debug_display ) {
          echo sprintf('Distance Break as d(%s) <= d(%s) from %s to %s', $min_dist, $ds, $from, $to) . '<br />';
        }
      }

      return $result;
    }

    function is_distance_segment_break($from, $via, $to, $current_array) {
      if( !$this->control_array['distance_segment'] ) return false;

      $min_dist = array_sum($current_array);

      foreach($this->path_result_array as $ck => $cv) {

        $tmp_array = explode('_', $ck);
        $ds = $tmp_array[0];

        if( $ds <= $min_dist && count($cv) <= count($current_array) ) {
          $this->stats_array['distance_segment']++;
          if( $this->debug_display ) {
            echo sprintf('Distance=%s Segment Break from %s via %s to %s', $min_dist, $from, $via, $to) . '<br />';
          }
          return true;
        }
        break;
      }
      return false;
    }

    function get_stats() {
      $result = '<br /><b>Statistics</b><br />';
      foreach( $this->stats_array as $key => $value ) {
        $result .= $key . ':&nbsp;<b>' . $value . '</b><br />';
      }
      return $result;
    }

    // Retrieve current optimum path
    function get_optimum() {
      $result_array = array();
      foreach($this->path_result_array as $ck => $cv) {
        $result_array[$ck] = $cv;
        break;
      }
      return $result_array;
    }

    function &get_result_paths() {
      return $this->path_result_array;
    }

    function get_result_strings() {
      ksort(nodes::$marked_array, SORT_NUMERIC);

      if( empty($this->path_result_array) ) {
        $result = sprintf('<br /><b>Cannot locate a path from %s to %s</b>', $this->org_from, $this->org_to) . '<br />';
        if( !empty($this->nearest) ) {
          $result .= sprintf('Nearest Node to %s found is %s with d(%s)', $this->org_to, $this->nearest['n'], $this->nearest['d']) . '<br />';
        }
      } else {
        $result = sprintf('<br /><b>Suggested Paths from %s to %s</b>', $this->org_from, $this->org_to) . '<br />';

        $index = 0;
        foreach($this->path_result_array as $key => $value) {
          $index++;
          $path_string = implode(',', array_keys($value));
          $result .= sprintf('%s. [%s] d:(%s)<br />', $index, $path_string, array_sum($value));
        }

        $result .= '<br/><b>Original Paths with order found during scan</b><br />';
        $index = 0;
        foreach($this->org_result_array as $key => $value) {
          $index++;
          $path_string = implode(',', array_keys($value));
          $result .= sprintf('%s. [%s] d:(%s)<br />', $index, $path_string, array_sum($value));
        }
      }


      if( $this->display_array['marked_array'] ) {
        $result .= '<br /><b>Cached Nodes Segments from attempted scans</b><br />';
        foreach(nodes::$marked_array as $key => $value) {
          $connects = '';
          //$path_string = implode(',', array_keys($value));
          foreach( $value as $ck => $cv ) {
            $tmp_array = array_keys($cv);
            if( !empty($tmp_array) ) {
              $connects .= $ck . ':[' . implode(',', $tmp_array) . ']=' . array_sum($cv) . ', ';
            } else {
              $connects .= $ck . ':[-], ';
            }
          }
          $connects = substr($connects, 0, -2);
          $result .= sprintf('%s -> %s<br />', $key, $connects);
        }
      }

      if( $this->display_array['nodes_array'] ) {
        $result .= '<br /><b>Nodes Matrix</b><br />';
        foreach(nodes::$nodes_array as $key => $value) {
          $path_string = implode(',', array_keys($value));
          $dist_string = implode(',', $value);
          $result .= sprintf('%s -> [%s] d(%s)<br />', $key, $path_string, $dist_string);
        }
      }

      if( $this->control_array['use_vertices'] && $this->display_array['verts_array'] ) {
        $result .= '<br /><b>Vertices Matrix</b><br />';
        foreach(nodes::$verts_array as $key => $value) {
          $d = nodes::get_node_ds($key, $this->org_to);
          $result .= sprintf('%s -> x:[%s] y:[%s] z[%s] d(%s)<br />', $key, $value['x'], $value['y'], $value['z'], $d);
        }
      }

      return $result;
    }

    static function load_path_file($filename) {

      $sep = "\r\n";
      $contents = '';
      $result = tep_read_contents($filename, $contents);
      if( !$result ) {
        echo 'Could not read node file<br />';
        return $result;
      }

      nodes::reset_statics();

      $contents_array = explode($sep.$sep, $contents);
      $tmp_array = explode($sep, $contents_array[0]);

      foreach($tmp_array as $line) {
        $nodes_array = unserialize(base64_decode($line));
        if( !is_array($nodes_array) ) {
          echo sprintf('Invalid Nodes detected - skipping<br />%s', var_dump($nodes_array) );
          continue;
        }
        nodes::$nodes_array += $nodes_array;
      }
      if( isset($contents_array[1]) ) {
        $tmp_array = explode($sep, $contents_array[1]);
        foreach($tmp_array as $line) {
          $verts_array = unserialize(base64_decode($line));
          if( !is_array($verts_array) ) {
            echo sprintf('Invalid Vertices detected - skipping<br />%s', var_dump($verts_array) );
            continue;
          }
          nodes::$verts_array += $verts_array;
        }
      }
      return $result;
    }

    static function save_path_file($filename) {
      $contents = '';
      $sep = "\r\n";

      $breaker = 64;
      $index = 0;
      $tmp_array = array();

      foreach(nodes::$nodes_array as $key => $value) {
        $index++;
        $tmp_array[$key] = $value;
        if( $index >= $breaker ) {
          $contents .= base64_encode(serialize($tmp_array)) . $sep;
          $index = 0;
          $tmp_array = array();
        }
      }

      if( !empty($tmp_array) ) {
        $contents .= base64_encode(serialize($tmp_array));
      } else {
        $contents = substr($contents, 0, -strlen($sep));
      }

      $index = 0;
      $tmp_array = array();
      if( !empty(nodes::$verts_array) ) {
        $contents .= $sep . $sep;
        foreach(nodes::$verts_array as $key => $value) {
          $index++;
          $tmp_array[$key] = $value;
          if( $index >= $breaker ) {
            $contents .= base64_encode(serialize($tmp_array)) . $sep;
            $index = 0;
            $tmp_array = array();
          }
        }

        if( !empty($tmp_array) ) {
          $contents .= base64_encode(serialize($tmp_array));
        } else {
          $contents = substr($contents, 0, -strlen($sep));
        }
      }

      $result = tep_write_contents($filename, $contents);
      if( !$result ) {
        echo 'Could not write node file<br />';
      }
      return $result;
    }

    static function create_random_nodes($max_nodes=100, $min_dist=1, $max_dist=8, $max_connects=4, $random_connect=true) {
      nodes::reset_statics();

      for($i=1, $j=$max_nodes; $i<=$j; $i++) {
        $rx = rand(1, min(100, $max_nodes) );
        $ry = rand(1, min(100, $max_nodes) );
        $rz = rand(1, min(100, $max_nodes) );
        nodes::$verts_array[$i] = array('x' => $rx, 'y' => $ry, 'z' => $rz);
      }

      for($i=1, $j=$max_nodes; $i<=$j; $i++) {
        $connects = $max_connects;
        if( $random_connect ) {
          $connects = rand(1, $max_connects);
        }
        for( $i2=0, $j2=$connects; $i2<$j2; $i2++ ) {
          do {
            $sub = rand(1, $max_nodes);
          //} while( isset(nodes::$nodes_array[$i][$sub]) || isset(nodes::$nodes_array[$sub][$i]) || ($i == $sub) );
          } while( isset(nodes::$nodes_array[$i][$sub]) || ($i == $sub) );

          $dist = nodes::get_node_ds($i, $sub, $min_dist, $max_dist);
          nodes::$nodes_array[$i][$sub] = $dist;
          //nodes::$nodes_array[$sub][$i] = $dist;
        }
      }

      nodes::short_nodes();
    }

    static function short_nodes() {
      ksort(nodes::$nodes_array, SORT_NUMERIC);
      foreach( nodes::$nodes_array as $key => $value) {
        asort(nodes::$nodes_array[$key], SORT_NUMERIC);
      }
    }
  }
?>
